/******************************************************
(c) 2002 - 2003 by Daniel Vollmer <maven@maven.de>.
Licensed under the GNU Public License V2.
*******************************************************/

texture tNormal;
texture tCube;
texture tRenderTarget;

float4x3	mWorld;
float3x3	mWorldIT;
float4x4	mWorldViewProj;

float4x4	mViewI; // for the SkyBox

float2		vTexCoordBias;

float3		vCamPos;
float3		vZZZ; // fMinZ, fOneOverMaxMinZ, fMaxZ - fMinZ
float3		vFilterColor;
float4		vEta; // fEta, 1 / fEta, fEta0, 1.0f - fEta0

sampler sRenderTarget = sampler_state
{
	Texture = <tRenderTarget>;
	MipFilter = NONE;
	MinFilter = POINT;
	MagFilter = POINT;
	AddressU = Clamp;
	AddressV = Clamp;
};

sampler sCube = sampler_state
{
	Texture = <tCube>;
	MipFilter = POINT;
	MinFilter = LINEAR;
	MagFilter = LINEAR;
	AddressU = Clamp;
	AddressV = Clamp;
	AddressW = Clamp;
};

sampler sNormal = sampler_state
{
	Texture = <tNormal>;
	MipFilter = LINEAR;
	MinFilter = ANISOTROPIC;
	MagFilter = LINEAR;
	MaxAnisotropy = 8;
};

struct VS_input
{
	float4 vPosition        : POSITION;
	float3 vNormal          : NORMAL;
	float2 vTexCoords       : TEXCOORD0;
	float3 vTangent         : TANGENT;
};

struct VS_output
{
	float4 vPosition	: POSITION;
	float2 vTexCoord0	: TEXCOORD0; // for normal-map, + depth
	float3x3 vTransform	: TEXCOORD1;
	float4 vWorldDir	: TEXCOORD4;
	float4 vScreenPos	: TEXCOORD5;
};

VS_output vs_rt_pass(const VS_input v)
{
	VS_output Output;

	float2 texpos;
	float3 normal, tangent, binormal;
	float3 dir;

	Output.vPosition = mul(v.vPosition, mWorldViewProj); // required transform to HPOS
	Output.vTexCoord0 = v.vTexCoords; // pass-through the normal map texture coords

	normal = mul(v.vNormal, mWorldIT); // transform the normal (for the tangent-space basis)
	tangent = mul(v.vTangent, mWorldIT); // transform the tangent (for the tangent-space basis)
	binormal = cross(normal, tangent); // create the binormal (for the tangent-space basis)

	// and put them into the matrix (which is used in the pixel shader)
	Output.vTransform[0] = float3(tangent.x, binormal.x, normal.x);
	Output.vTransform[1] = float3(tangent.y, binormal.y, normal.y);
	Output.vTransform[2] = float3(tangent.z, binormal.z, normal.z);

	dir = mul(v.vPosition, mWorld) - vCamPos; // and compute the world-space direction vector

	// and generate direction as well as distance (scaled to [0, 1] for the object)
	Output.vWorldDir = float4(dir, (length(dir) - vZZZ.x) * vZZZ.y);

	Output.vScreenPos = Output.vPosition;
	Output.vScreenPos.y = -Output.vScreenPos.y; // flip y-axis for the rendertarget reads

	return Output;
}

float4 ps_rt_pass(const VS_output v) : COLOR
{
	float3 N = tex2D(sNormal, v.vTexCoord0.xy)*2-1; // get the normal from the normal map

	N = mul(v.vTransform, N); // transform the normal by tangent-space-basis
	return float4(N*(-0.5)+0.5, v.vWorldDir.w); // store normal & depth
}

technique StoreBackFaceNormalPass
{
	pass P0
	{
		CullMode = CW;
//		LastPixel = TRUE;
		VertexShader = compile vs_1_1 vs_rt_pass();
		PixelShader = compile ps_2_0 ps_rt_pass();
	}
}

// this function get an HPOS vector and transforms that to a (u,v) coordinate to read from
// (as we can't read vPos in a 2.x pixel shader (only 3.0))
static float2 rt_project(float4 ScreenSpace)
{
	return ScreenSpace.xy / (2 * ScreenSpace.w) + vTexCoordBias;
}

// refraction of I at N with relative index of refraction eta
static float3 refr(float3 I, const float3 N, const float eta)
{
	I = -I;
	float NdotI = dot(N, I);

	return (eta * NdotI - sqrt((1 - eta*eta*(1-NdotI*NdotI)))) * N - eta * I;
}

static float fresnel_approx(const float3 dir, const float3 normal, uniform const float R0, uniform const float OneMR0)
{
	return R0 + OneMR0 * pow(saturate(1+dot(dir, normal)), 5);
}

float4 ps_final_pass(const VS_output v) : COLOR
{
	float3 N = tex2D(sNormal, v.vTexCoord0.xy)*2-1; // look up normal and convert from RGB
	float3 dir;
	float3 color_refr;
	float3 color_refl;
	float4 temp;
	float fr, depth;

	N = mul(v.vTransform, N); // transform the normal from the normal map by the tangent-space basis

	dir = normalize(v.vWorldDir.xyz); // normalize the world-space direction vector between camera and the pixel
	color_refl = texCUBE(sCube, reflect(dir, N)); // look up the reflection of the current pixel
	
	fr = fresnel_approx(dir, N, vEta.z, vEta.w);

	dir = refr(dir, N, vEta.x); // then refract the direction vector *into* the cube
	temp = tex2D(sRenderTarget, rt_project(v.vScreenPos)); // and get the normal for the back-face of the cube
	N = temp.xyz*2-1; // expand the normal from RGB -> [-1, 1]
	dir = refr(dir, N, vEta.y); // and refract direction *out* of the cube (vEta.y = 1 / vEta.x)
	color_refr = texCUBE(sCube, dir); // and finally get the color of the twice refracted ray

	depth = (temp.w - v.vWorldDir.w) * vZZZ.z; // convert the depth back to world-space units
	// and filter out the transmitted colour according to depth through the cube
	color_refr = saturate(color_refr * (1 - depth * vFilterColor));

	// now blend between refracted and reflected light based on fresnel approximation
	return float4(lerp(color_refr, color_refl, fr), 1);
}

technique SkyBoxPass
{
	pass P0
	{ // draws the skybox (using the fixed function pipeline)
		CullMode = CW; // we want to draw the cube's interior
		Ambient = 0xffffffff;
		Lighting = FALSE;
		VertexShader = NULL;
		PixelShader = NULL;
		Sampler[0] = <sCube>;
		TextureTransformFlags[0] = COUNT3;
		TexCoordIndex[0] = CAMERASPACENORMAL;
		TextureTransform[0] = <mViewI>;
	}
}

technique FinalPass
{
	pass P0
	{
		CullMode = CCW;
		Wrap0 = 0; // needed for non-POW2 textures
		VertexShader = compile vs_1_1 vs_rt_pass();
		PixelShader = compile ps_2_0 ps_final_pass();
	}
}